/*
 *  RSA_encrypt.c
 *  
 *  Author  : Ahmed Obied (ahmed.obied@gmail.com)
 *  Version : 0.1
 *
 *  This program implements the RSA cryptosystem for encryption and descryption
 *
 */

#include <cstdlib>
#include <cstdio>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/procfs.h>
#include <sys/mman.h>
#include <NTL/ZZ.h>
#include "RSA_encrypt.h"

#define	N_BITS		1024
#define PRIME_TEST	100
#define BIN_BASE	2
#define DECIMAL_BASE	10
#define	P_BLOCK_SIZE	(N_BITS/8) - 1
#define	C_BLOCK_SIZE	(N_BITS/8)
#define MAX_CMD		30
#define STARTING_E	65537

RSA_encrypt::RSA_encrypt()
{
	mpz_init(e);
	mpz_init(d);
	mpz_init(n);
}

RSA_encrypt::~RSA_encrypt()
{
	mpz_clear(e);
	mpz_clear(d);
	mpz_clear(n);
}

void RSA_encrypt::initializeKeys()
{
	int fd, err;
	mpz_t p, q, phi, tmp;
	gmp_randstate_t rand_stat;
	unsigned long int rand_seed;
	// Use the /dev/random to get a seed value
	fd = open("/dev/random",O_RDONLY);
	if (fd < 0) {
		perror("initializeKeys: open");
		exit(-1);
	}
	err = read(fd,&rand_seed,sizeof(unsigned long int));
	if (err < 0) {
		perror("initializeKeys: read");
		exit(-1);
	}
	// initialize the random number generator
	gmp_randinit(rand_stat,GMP_RAND_ALG_LC);
	mpz_init(p);
	mpz_init(q);
	mpz_init(tmp);
	// seed the random number generator
	gmp_randseed_ui(rand_stat,rand_seed);
	// we keep searching for strong primes
	while(1) {
		mpz_set_ui(p,0);
		mpz_set_ui(q,0);
		// generate prime p with 512-bits
		while (!mpz_probab_prime_p(p,PRIME_TEST)) {
			// select a random number
			mpz_urandomb(p,rand_stat,(N_BITS/2));
			/* set the 2 high-order bits to force the number
			 * to have the exact number of bits passed
			 */
			if (!mpz_tstbit(p,(N_BITS/2) - 1))
				mpz_setbit(p,(N_BITS/2) - 1);
			if (!mpz_tstbit(p,(N_BITS/2) - 2))
				mpz_setbit(p,(N_BITS/2) - 2);
			// set low-order bit to force the number to be odd
			if (!mpz_tstbit(p,0))
				mpz_setbit(p,0);
		}
		// generate prime q with 512 bits
		while (!mpz_probab_prime_p(q,PRIME_TEST)) {
			// select a random number first
			mpz_urandomb(q,rand_stat,(N_BITS/2));
			// we need the exact number of bits
			if (!mpz_tstbit(q,(N_BITS/2) - 1))
				mpz_setbit(q,(N_BITS/2) - 1);
			if (!mpz_tstbit(q,(N_BITS/2) - 2))
				mpz_setbit(q,(N_BITS/2) - 2);
			// a prime cannot be even
			if (!mpz_tstbit(q,0))
				mpz_setbit(q,0);
		}
		// if p = q then we re-generate
		if (!mpz_cmp(p,q))
			continue;
		// computer n = pq which is 1024-bits
		mpz_mul(n,p,q);
		// if |p - q| < n^(1/4) then we re-generate
		mpz_sub(tmp,p,q);
		mpz_abs(tmp,tmp);
		mpz_root(e,n,4);
		if (mpz_cmp(tmp,e) < 0)
			continue;
		break;
	}
	// compute phi(n) = (p - 1)(q - 1)
	mpz_init(phi);
	mpz_sub_ui(p,p,1);
	mpz_sub_ui(q,q,1);
	mpz_mul(phi,p,q);
	// start with e = 65537 and increment by 2
	mpz_set_ui(e,STARTING_E);
	// gcd(e,phi(n)) must be 1 for the process to work
	while (1) {
		mpz_gcd(tmp,e,phi);
		// gcd(e,phi(n)) = 1 , done
		if (!mpz_cmp_ui(tmp,1))
			break;
		mpz_add_ui(e,e,2);
	}
	// compute d = e^{-1} mod phi(n)
	mpz_invert(d,e,phi);
	// clear some memory
	mpz_clear(p);
	mpz_clear(q);
	mpz_clear(phi);
	mpz_clear(tmp);
}

void RSA_encrypt::encrypt_file(char *infile, char *outfile)
{
	int in_fd, out_fd, err;
	char *p_data, *c_data;
	struct stat in_stat, out_stat;
	bool fin_flag;
	ZZ zz_e, zz_n, num_msg, enc_msg;

	/* if the source file for plaintext and the file where we want to
         * are the same, then output an error ...
	 */
	if (strcmp(infile,outfile) == 0) {
		fprintf(stderr,"encrypt_file: Can't overwrite source file\n");
		exit(-1);
	}
	// open plaintext message
	in_fd = open(infile,O_RDONLY);
	if (in_fd < 0) {
		perror("ecnrypt_file: open infile");
		exit(-1);
	}
	// open a file to write the ciphertext to it
	out_fd = open(outfile,O_WRONLY|O_CREAT);
	if (out_fd < 0) {
		perror("encrypt_file: open outfile");
		exit(-1);
	}
	// get some info from file e.g. size, etc
	err = fstat(in_fd,&in_stat);
	if (err < 0) {
		perror("encrypt_file: fstat");
		exit(-1);
	}
	err = fstat(out_fd,&out_stat);
	if (err < 0) {
		perror("encrypt_file: fstat outfile");
		exit(-1);
	}
	// setting the mode in case the file was created
	err  = fchmod(out_fd,S_IRUSR|S_IWUSR);
	if (err < 0) {
		perror("encrypt_file: fchmod");
		exit(-1);
	}
	// if the output file is not empty then we /dev/null it
	if (out_stat.st_size != 0) {
		char* cmd = (char *)malloc(MAX_CMD);
		char* partial_cmd = "cat /dev/null > ";
		strcpy(cmd,partial_cmd);
		strncat(cmd,outfile,(size_t)(MAX_CMD - strlen(partial_cmd)));
		// execute system to clear the contents of the output file
		err = system(cmd);
		if (err < 0) {
			perror("encrypt_file: system");
			exit(-1);
		}
		free(cmd);
	}
	size = in_stat.st_size;
	/*  create plaintext and ciphertext blocks to allow faster encryption
	 *  plaintext block is 127 bytes and ciphertext block is 128 bytes
         */
	p_data = (char *)malloc(P_BLOCK_SIZE);
	c_data = (char *)malloc(C_BLOCK_SIZE);
	// from here we start working with the NTL library
	zz_n = to_ZZ(mpz_get_str(NULL,DECIMAL_BASE,n));
	zz_e = to_ZZ(mpz_get_str(NULL,DECIMAL_BASE,e));
	fin_flag = false;
	// keep reading until we don't have any more blocks to read
	while (1) {
		int bytes_read = 0;

		// initialize plaintext block to avoid problems
		memset(p_data,0,P_BLOCK_SIZE);
		// read a block of 127 bytes
		bytes_read = read(in_fd,p_data,P_BLOCK_SIZE);
		if (bytes_read < 0) {
			perror("encrypt_file: read");
			exit(-1);
		}
		if (bytes_read < P_BLOCK_SIZE)
			fin_flag = true;
		// represent the block as an integer <= n - 1
		num_msg = ZZFromBytes((const unsigned char *)p_data,bytes_read);
		// RSA encrypt algorithm cipher = M^e mod n
		enc_msg = PowerMod(num_msg,zz_e,zz_n);
		// set ciphertext to zero to padd automatically
		memset(c_data,0,C_BLOCK_SIZE);
		// convert to bytes
		BytesFromZZ((unsigned char *)c_data,enc_msg,C_BLOCK_SIZE);
		// write the bytes to the file ... its going to be raw binary
		err = write(out_fd,c_data,C_BLOCK_SIZE);
		if (err < 0) {
			perror("encrypt_file: write");
			exit(-1);
		}
		if (fin_flag)
			break;
	}
	// clear some memroy
	num_msg.kill();
	enc_msg.kill();
	zz_n.kill();
	zz_e.kill();
	free(p_data);
	free(c_data);
	close(in_fd);
	close(out_fd);
}

void RSA_encrypt::decrypt_file(char *infile, char *outfile)
{
	int in_fd, out_fd, err, c_size, p_size;
	char *p_data, *c_data;
	struct stat in_stat, out_stat;
	ZZ zz_d, zz_n, num_msg, dec_msg;

	/* if the source file for ciphertext and the file where we want to
         * are the same, then output an error ...
	 */
	if (strcmp(infile,outfile) == 0) {
		fprintf(stderr,"decrypt_file: Can't overwrite source file\n");
		exit(-1);
	}
	// open encrypted file
	in_fd = open(infile,O_RDONLY);
	if (in_fd < 0) {
		perror("decrypt_file: open infile");
		exit(-1);
	}
	// if output file is not there then create it
	out_fd = open(outfile,O_WRONLY|O_CREAT);
	if (out_fd < 0) {
		perror("decrypt_file: open outfile");
		exit(-1);
	}
	err = fstat(in_fd,&in_stat);
	if (err < 0) {
		perror("decrypt_file: fstat infile");
		exit(-1);
	}
	err = fstat(out_fd,&out_stat);
	if (err < 0 ) {
		perror("decrypt_file: fstat outfile");
		exit(-1);
	}
	// set modes to be able to open the file after creating
	err  = fchmod(out_fd,S_IRUSR|S_IWUSR);
	if (err < 0) {
		perror("decrypt_file: fchmod");
		exit(-1);
	}
	// if output file is not empty then clear it
	if (out_stat.st_size != 0) {
		char* cmd = (char *)malloc(MAX_CMD);
		char* partial_cmd = "cat /dev/null > ";
		strcpy(cmd,partial_cmd);
		strncat(cmd,outfile,(size_t)(MAX_CMD - strlen(partial_cmd)));
		err = system(cmd);
		if (err < 0) {
			perror("decrypt_file: system");
			exit(-1);
		}
		free(cmd);
	}
	// create plaintext and ciphertext blocks
	p_data = (char *)malloc(P_BLOCK_SIZE);
	c_data = (char *)malloc(C_BLOCK_SIZE);
	// again converting to NTL library
	zz_n = to_ZZ(mpz_get_str(NULL,DECIMAL_BASE,n));
	zz_d = to_ZZ(mpz_get_str(NULL,DECIMAL_BASE,d));
	c_size = 0;
	p_size = 0;
	while (1) {
		int bytes_read = 0;
		
		memset(c_data,0,C_BLOCK_SIZE);
		// read ciphertext block ... 128 bytes each
		bytes_read = read(in_fd,c_data,C_BLOCK_SIZE);
		if (bytes_read < 0) {
			perror("decrypt_file: read");
			exit(-1);
		}
		if (bytes_read == 0)
			break;
		c_size += C_BLOCK_SIZE;
		// convert the ciphertext block to an integer <= n - 1
		num_msg = ZZFromBytes((const unsigned char *)c_data,bytes_read);
		// RSA decryption M = C^d mod n
		dec_msg = PowerMod(num_msg,zz_d,zz_n);
		memset(p_data,0,P_BLOCK_SIZE);
		// get the original bytes
		BytesFromZZ((unsigned char *)p_data,dec_msg,P_BLOCK_SIZE);
		// calculate correct offset for the last block
		if (c_size == in_stat.st_size) {
			int l_block = 0;
			// check out the correct number of bytes
			p_size += NumBytes(dec_msg);
			while (1) {
				if (p_size == size)
					break;
				p_size++;
				l_block++;
			}
			/* last block is treated differently to here is where
                         * we write it to the file
			 */
			err = write(out_fd,p_data,NumBytes(dec_msg) + l_block);
			if (err < 0) {
				perror("decrypt_file: write");
				exit(-1);
			}
		}
		else {
			// write the plaintext block which is 127 bytes to the file
			err = write(out_fd,p_data,P_BLOCK_SIZE);
			if (err < 0) {
				perror("decrypt_file: write");
				exit(-1);
			}
			p_size += P_BLOCK_SIZE;
		}
	}
	// free memory
	zz_n.kill();
	zz_d.kill();
	dec_msg.kill();
	num_msg.kill();
	free(p_data);
	free(c_data);
	close(in_fd);
	close(out_fd);
}

