/*
 *  RSA_signature.c
 *
 *  Author  : Ahmed Obied (ahmed.obied@gmail.com)
 *  Version : 0.1
 *
 *  This program implements the RSA digital signature scheme
 *
 */

#include <cstdlib>
#include <fcntl.h>
#include <string.h>
#include "RSA_signature.h"
#include "sha1.h"

#define	N_BITS		1024
#define BIN_BASE	2
#define DECIMAL_BASE	10
#define	BASE		256
#define PRIME_TEST	100
#define C_BLOCK_SIZE	128
#define STARTING_E	65357

RSA_signature::RSA_signature()
{
	mpz_init(e);
	mpz_init(d);
	mpz_init(n);
}

RSA_signature::~RSA_signature()
{
	mpz_clear(e);
	mpz_clear(d);
	mpz_clear(n);
}

void RSA_signature::initializeKeys()
{
	int fd, err;
	mpz_t p, q, phi, tmp;
	gmp_randstate_t rand_stat;
	unsigned long int rand_seed;

	// Use the /dev/random to get a seed value
	fd = open("/dev/random",O_RDONLY);
	if (fd < 0) {
		perror("initializeKeys: open");
		exit(-1);
	}
	err = read(fd,&rand_seed,sizeof(unsigned long int));
	if (err < 0) {
		perror("initializeKeys: read");
		exit(-1);
	}
	// initialize the random number generator
	gmp_randinit(rand_stat,GMP_RAND_ALG_LC);
	mpz_init(p);
	mpz_init(q);
	mpz_init(tmp);
	// seed the random number generator
	gmp_randseed_ui(rand_stat,rand_seed);
	// we keep searching for strong primes
	while(1) {
		mpz_set_ui(p,0);
		mpz_set_ui(q,0);
		// generate prime p with 512-bits
		while (!mpz_probab_prime_p(p,PRIME_TEST)) {
			// select a random number
			mpz_urandomb(p,rand_stat,(N_BITS/2));
			/* set the 2 high-order bits to force the number
			 * to have the exact number of bits passed
			 */
			if (!mpz_tstbit(p,(N_BITS/2) - 1))
				mpz_setbit(p,(N_BITS/2) - 1);
			if (!mpz_tstbit(p,(N_BITS/2) - 2))
				mpz_setbit(p,(N_BITS/2) - 2);
			// set low-order bit to force the number to be odd
			if (!mpz_tstbit(p,0))
				mpz_setbit(p,0);
		}
		// generate prime q with 512 bits
		while (!mpz_probab_prime_p(q,PRIME_TEST)) {
			// select a random number first
			mpz_urandomb(q,rand_stat,(N_BITS/2));
			// we need the exact number of bits
			if (!mpz_tstbit(q,(N_BITS/2) - 1))
				mpz_setbit(q,(N_BITS/2) - 1);
			if (!mpz_tstbit(q,(N_BITS/2) - 2))
				mpz_setbit(q,(N_BITS/2) - 2);
			// a prime cannot be even
			if (!mpz_tstbit(q,0))
				mpz_setbit(q,0);
		}
		// if p = q then we re-generate
		if (!mpz_cmp(p,q))
			continue;
		// computer n = pq which is 1024-bits
		mpz_mul(n,p,q);
		// if |p - q| < n^(1/4) then we re-generate
		mpz_sub(tmp,p,q);
		mpz_abs(tmp,tmp);
		mpz_root(e,n,4);
		if (mpz_cmp(tmp,e) < 0)
			continue;
		break;
	}
	// compute phi(n) = (p - 1)(q - 1)
	mpz_init(phi);
	mpz_sub_ui(p,p,1);
	mpz_sub_ui(q,q,1);
	mpz_mul(phi,p,q);
	// start with e = 65537 and increment by 2
	mpz_set_ui(e,STARTING_E);
	// gcd(e,phi(n)) must be 1 for the process to work
	while (1) {
		mpz_gcd(tmp,e,phi);
		// gcd(e,phi(n)) = 1 , done
		if (!mpz_cmp_ui(tmp,1))
			break;
		mpz_add_ui(e,e,2);
	}
	// compute d = e^{-1} mod phi(n)
	mpz_invert(d,e,phi);
	// clear some memory
	mpz_clear(p);
	mpz_clear(q);
	mpz_clear(phi);
	mpz_clear(tmp);
}

unsigned int RSA_signature::sign_file(char *infile, unsigned char **signature)
{
	int in_fd, err;
	struct stat in_stat;
	SHA1Context sha1;
	uint8_t Message_Digest[20];
	unsigned char* sig;
	unsigned int i;
	mpz_t num_sig, enc_sig;
	size_t size;

	// open file to be signed
	in_fd = open(infile,O_RDONLY);
	if (in_fd < 0) {
		perror("sign_file: open");
		exit(-1);
	}
	// get size of file to read it all
	fstat(in_fd,&in_stat);
	char msg[in_stat.st_size];
	// read all the bytes in the file
	err = read(in_fd,msg,in_stat.st_size);
	if (err < 0) {
		perror("sign_file: read");
		exit(-1);
	}
	// initialize the sha1Context struct
	err = SHA1Reset(&sha1);
	if (err) {
		fprintf(stderr,"sign_file: sha1 reset error\n");
		exit(1);
	}
	// divide the file into 512-bits and add pads if needed
	err = SHA1Input(&sha1,(const unsigned char *)msg,in_stat.st_size);
	if (err) {
		fprintf(stderr,"sign_file: sha1 input error\n");
		exit(1);
	}
	// computer the hash value
	err = SHA1Result(&sha1, Message_Digest);
        if (err) {
        	fprintf(stderr,"sign_file: sha1 result error\n");
        	exit(1);
        }
	sig = Message_Digest;
	mpz_init(num_sig);
	// represent a sequence of bytes as big integer
	for (i = 0 ; i < strlen((const char *)sig) ; i++) {
		unsigned long int ascii_code;

		ascii_code = (unsigned long int)sig[i];
		mpz_add_ui(num_sig,num_sig,ascii_code);
		mpz_mul_ui(num_sig,num_sig,BASE);
	}
	mpz_add_ui(num_sig,num_sig,(unsigned long int)strlen((const char *)sig));
	mpz_init(enc_sig);
	// use private key d to sign
	mpz_powm(enc_sig,num_sig,d,n);
	// allocate memory for the signature
	*signature = (unsigned char *)malloc(C_BLOCK_SIZE);
	memset(*signature,0,C_BLOCK_SIZE);
	size = 0;
	// copy from mpz to bytes which gives us raw binary
	mpz_export(*signature,&size,1,C_BLOCK_SIZE,0,0,enc_sig);
	mpz_clear(enc_sig);
	mpz_clear(num_sig);
	// return size of signature
	return C_BLOCK_SIZE;
}

bool RSA_signature::verify_file(char *infile,unsigned char *signature)
{
	int in_fd, err, i, sig_len;
	struct stat in_stat;
	SHA1Context sha1;
	uint8_t Message_Digest[20];
	unsigned char *sig, *dec_sig;
	mpz_t num_sig, enc_sig, tmp;
	bool sig_test;

	// same as sign_file
	in_fd = open(infile,O_RDONLY);
	if (in_fd < 0) {
		perror("verify_file: open");
		exit(-1);
	}
	fstat(in_fd,&in_stat);
	char msg[in_stat.st_size];

	err = read(in_fd,msg,in_stat.st_size);
	if (err < 0) {
		perror("verify_file: read");
		exit(-1);
	}
	err = SHA1Reset(&sha1);
	if (err) {
		fprintf(stderr,"verify_file: sha1 reset error\n");
		exit(1);
	}
	err = SHA1Input(&sha1,(const unsigned char *)msg,in_stat.st_size);
	if (err) {
		fprintf(stderr,"verify_file: sha1 input error\n");
		exit(1);
	}
	err = SHA1Result(&sha1, Message_Digest);
        if (err) {
        	fprintf(stderr,"verify_file: sha1 result error\n");
        	exit(1);
        }
	// sig now has the hash value of the file that was passed
	sig = Message_Digest;
	mpz_init(num_sig);
	mpz_init(enc_sig);
	// covert the bytes to an mpz_t type
	mpz_import(enc_sig,1,1,C_BLOCK_SIZE,0,0,signature);
	mpz_powm(num_sig,enc_sig,e,n);
	mpz_init(tmp);
	// starting reversing the string to big integer process
	mpz_mod_ui(tmp,num_sig,BASE);
	sig_len = mpz_get_ui(tmp);
	// dec_sig will have the signature that we get
	dec_sig = (unsigned char *)malloc(sig_len);
	mpz_sub_ui(num_sig,num_sig,sig_len);
	for (i = sig_len - 1 ; i >= 0 ; i--) {
		int curr_ascii;
		mpz_init(tmp);
		// div, mod to current original ascii
		mpz_div_ui(num_sig,num_sig,BASE);
		mpz_mod_ui(tmp,num_sig,BASE);
		curr_ascii = mpz_get_ui(tmp);
		// get original bytes
		dec_sig[i] = (unsigned char)curr_ascii;
	}
	sig_test = true;
	// test whether the two signatures are the same or not
	for (i = 0 ; i < (signed int)strlen((const char *)sig) ; i++) {
		if (sig[i] != dec_sig[i]) {
			// not a valid signature
			sig_test = false;
			break;
		}
	}
	free(dec_sig);
	mpz_clear(num_sig);
	mpz_clear(enc_sig);
	mpz_clear(tmp);
	// if signature is valid return true and false otherwise
	return sig_test;
}
